from urlparse import urlparse

from s3 import S3Storage
from cloudfiles import CloudfilesStorage

available_services = {
    's3': S3Storage,
    # Cloudfiles currently disabled, it's just not production ready yet.
    # 'cf': CloudfilesStorage
}


def luggage(url, credentials=()):
    """
    Creates a handle to the cloud storage denoted by the URL.

    The following URL formats are supported:

    * Amazon S3: s3://[region]/bucket/ with credentials=(access_key, secret_key)
    * Cloudfiles: cf://region/container/ with credentials=(username, key)
    """
    parsed_url = urlparse(url)

    try:
        service = available_services[parsed_url.scheme]
        return service(parsed_url, credentials)
    except KeyError:
        raise ValueError('Unsupported service {}'.format(parsed_url.scheme))


def get_available_schemes():
    return available_services.keys()


import exceptions


class ObjectInfo:
    """
    Provides information and operations on an object on a cloud storage.
    """

    def __init__(self, url, path, mtime, size, etag, extra, storage=None):
        self.url = url
        self.path = path
        self.mtime = mtime
        self.size = size
        self.etag = etag
        self.extra = extra
        self.storage = storage

    def __str__(self):
        return self.url
    __repr__ = __str__

    def delete(self):
        """
        Deletes this object from the storage.
        """
        try:
            self.storage.delete(self.path)
        except exceptions.ObjectDoesNotExistException:
            pass  # If the object didn't exist in the first place we don't need to do anything.

    def get(self, destination=''):
        """
        Downloads the object to the destination.

        :param destination: The destination on the local filesystem. If not set, the path of this object is used
        """
        self.storage.pull(self.path, destination)

    def __eq__(self, other):
        try:
            return self.url == other.url
        except AttributeError:
            return False


class CloudStorage:
    """
    The parent class for all storage implementations.

    A "storage" is a service on the web which allows remote storage in a file-like manner.
    These storages may be divided into multiple "containers" which can vary in accessibility and size.
    These containers then contain "objects" which are the file-like objects on the remote machine.

    Or in short: Storage (S3, CloudFiles, Dropbox) > Container (Bucket, Container, Box) > Object (or Key, i.e. test.txt)

    The methods in this class are the ones provided by all storage services. Those storage classes
    may offer additional methods, consult their documentation.
    """

    def __init__(self, url):
        self.url = url

    def _unimplemented(self, method):
        message = 'CloudStorage {0} has not implemented the "{1}" method'.format(self.__class__.__name__, method)
        return AttributeError(message)

    def list(self, path=''):
        """
        Lists object names under a given path.

        :param path: The path to search under, or an empty string if you want to search in the root of the storage
        :returns: A list of object names (i.e. a list of str)
        """
        raise self._unimplemented('list')

    def find(self, obj):
        """
        Retrieves the ObjectInfo for a single object.

        :param object: The path relative to the root of the storage
        :returns: ObjectInfo
        """
        raise self._unimplemented('find')

    def findall(self, path=''):
        """
        Finds all objects with complete meta-information.

        This will be a costly operation to do on most cloud storage services.

        :param path: The path to search under, or an empty string if you want to search in the root of the storage
        :returns: A list containing the ObjectInfo for each object
        """
        raise self._unimplemented('lsall')

    def delete(self, obj):
        """
        Deletes an object from the storage

        :param obj: The path which to delete
        :returns: True if this request deleted the file, false in all other cases
        """
        raise self._unimplemented('delete')

    def put_file(self, local_file, remote_path=None, verify=False):
        """
        Pushes the local file to a remote path.

        :param local_file: The local file to transfer
        :param remote_path: The path on the remote side. If not set, the basename of the local file is used.
        :param verify: Whether or not the successful transfer of the file should be verified after transmission (by md5)
        """
        raise self._unimplemented('push')

    def put_data(self, data, remote_path, verify=False):
        """
        Pushes in-memory data to a remote path.

        :param data: The data to push
        :type data: str or bytes
        :param remote_path: The path on the remote side.
        :param verify: Whether or not the successful transfer of the file should be verified after transmission (by md5)
        """
        raise self._unimplemented('push_data')

    def get(self, remote_path, local_file=None):
        """
        Pulls from the remote path to a local file.

        If the local file is not set, the full path of the remote file will be used, relative to the current
        working directory.
        """
        raise self._unimplemented('pull')

    def read(self, obj, binary=False):
        """
        Reads out the contents of the given object.

        :param obj: Object path
        :param binary: If set to true, a bytes-object will be returned instead of a string
        """
        raise self._unimplemented('read')

    def exists(self, obj):
        """
        Checks whether or not the given object exists.
        """
        raise self._unimplemented('exists')



from distutils.core import setup
from setuptools.command.test import test as TestCommand
import sys


# The whole Tox business was taken from "Open Sourcing a Python Project the Right Way"
class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)

setup(
    name='luggage',
    version='0.1.2b',
    description='A wrapper library for various cloud storage services',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
    ],
    author='Patrick Stegmann (aka. @wonderb0lt)',
    author_email='code@patrick-stegmann.de',
    url='https://github.com/wonderb0lt/luggage',
    packages=['luggage'],
    install_requires=['requests', 'python-dateutil'],
    tests_require=['tox'],
    cmdclass={'test': Tox}
)



import exceptions
import json
from datetime import datetime

import dateutil.parser as dateparser
import requests

import interface


class CloudfilesStorage(interface.CloudStorage):
    AUTH_ENDPOINT = 'https://auth.api.rackspacecloud.com/v1.1/auth'

    def __init__(self, url, credentials):
        interface.CloudStorage.__init__(self, url.geturl())
        self.region = str(url.netloc).upper()
        self.container = str(url.path).strip('/')
        self.mode = url.query.params.get('mode', 'public')
        self.user, self.key = credentials
        self.token = self.endpoint = self.token_expiry = None

    def auth(self):
        if not self.token_expiry or self.token_expiry < datetime.now():
            headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
            payload = {'credentials': { 'username': self.user, 'key': self.key}}
            r = requests.post(self.AUTH_ENDPOINT, headers=headers, data=json.dumps(payload))

            if r.status_code / 100 == 2:
                response = r.json()
                self.token = response['auth']['token']['id']
                self.token_expiry = dateparser.parse(response['auth']['token']['expires'])
                for region in response['auth']['serviceCatalog']['cloudFiles']:
                    if region['region'] == self.region:
                        self.endpoint = region[self.mode + 'URL'].replace('\\', '')
                        break
                else:
                    raise Exception('Unknown region: {}'.format(self.region))
            else:
                # TODO: More detailed error message
                raise exceptions.InvalidCredentialsException('Auth against rackspace failed')

    def container_url(self):
        return self.endpoint + '/' + self.container

    def object_url(self, obj):
        return self.container_url() + '/' + obj

    def request(self, method, url, **kwargs):
        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['X-Auth-Token'] = self.token
        return requests.request(method, url, **kwargs)

    def ls(self, path=''):
        self.auth()
        r = self.request('GET', self.container_url())

        object_count = int(r.headers['X-Container-Object-Count'])
        grepped = (r.text.split('\n')[:-1])  # Take away trailing newline

        while len(grepped) < object_count:
            r = self.request('GET', self.container_url(), params={'marker': grepped[-1]})
            grepped += (r.text.split('\n')[:-1])
            object_count = int(r.headers['X-Container-Object-Count'])

        return grepped



from os import makedirs
from os.path import basename, abspath, split, exists
from base64 import b64encode
from urlparse import urlparse

import requests
from dateutil.parser import parse as parsedate

import exceptions
import interface
import util

import xml.etree.ElementTree as ET


def f(node, tag):
    """
    Shorthand for node.find(amazon_ns + tag).text
    """
    n = node.find('{http://s3.amazonaws.com/doc/2006-03-01/}' + tag)
    return n.text if n is not None else ''


def fall(node, tag):
    """
    Shorthand for node.findall(amazon_ns + tag)
    """
    return node.findall('{http://s3.amazonaws.com/doc/2006-03-01/}' + tag)


def parse_bucket_listing(data):
    """
    Takes a ListBucketResult and transforms it to JSON.
    """
    r = {'contents': []}
    root = ET.fromstring(data)

    r['truncated'] = f(root, 'IsTruncated') == 'true'
    if r['truncated']:
        r['next'] = f(root, 'NextMarker')

    for key in fall(root, 'Contents'):
        r['contents'].append({
            'name': f(key, 'Key'),
            'last_modified': f(key, 'LastModified'),
            'etag': f(key, 'ETag').strip('"'),
            'size': f(key, 'Size')
        })

    return r



class S3Storage(interface.CloudStorage):
    def __init__(self, url, credentials):
        interface.CloudStorage.__init__(self, url.geturl())
        self.bucket = url.netloc
        self.key_id, self.secret_access_key = credentials

        self.session = requests.Session()
        r = self.request('HEAD', '/')

        if r.status_code == 404:
            raise exceptions.ContainerDoesNotExistException('Bucket {0} does not exist'.format(self.bucket))
        elif r.status_code == 403:
            raise exceptions.InvalidCredentialsException('The provided credentials for key {0} '
                                                         'are invalid'.format(self.key_id))
        elif r.status_code != 200:
            raise exceptions.LuggageException.from_respone(r)

    def request(self, verb, path, stream=False, **kwargs):
        url = 'https://{0}.s3.amazonaws.com/{1}'.format(self.bucket, path.lstrip('/'))

        if 'headers' in kwargs:
            kwargs['headers'].update({'Date': util.rfc1123_date()})
        else:
            kwargs['headers'] = {'Date': util.rfc1123_date()}

        request = self.sign_request(requests.Request(verb, url, **kwargs))
        request.stream = True
        return self.session.send(request, stream=stream)

    def sign_request(self, request):
        """
        Prepares and signs a requests.Request

        :param request: The request to sign
        :type request: requests.Request
        :returns: The prepared request, with the Authorization header set
        """
        prepared = request.prepare()
        verb = prepared.method
        content_type = prepared.headers.get('Content-Type', '')
        md5 = ''  # MD5 is optional and caused errors, not using it for now
        # md5 = util.md5(prepared.body)
        date = prepared.headers.get('Date', util.rfc1123_date())
        canonicalized_resource = '/' + self.bucket + urlparse(request.url).path

        canonicalized_headers = ''
        for amz_header in sorted([h for h in request.headers if h.startswith('X-Amz-')]):
            canonicalized_headers += '{0}:{1}\n'.format(amz_header.lower(), request.headers[amz_header])

        to_sign = '{verb}\n' \
                  '{md5}\n' \
                  '{content_type}\n' \
                  '{date}\n' \
                  '{canonicalized_headers}{canonicalized_resource}'.format(**locals())
        print to_sign
        signature = b64encode(util.hmac_sha1(key=self.secret_access_key, to_sign=to_sign))
        prepared.headers['Authorization'] = 'AWS {0}:{1}'.format(self.key_id, signature)
        return prepared

    def _to_info(self, key):
        url = 's3://{0}/{1}'.format(self.bucket, key['name'])
        path = key['name']
        mtime = parsedate(key['last_modified'])
        etag = key['etag']
        size = int(key['size'])
        return interface.ObjectInfo(url, path, mtime, size, etag, {}, self)

    def _infos_from_keys(self, keys):
        return [self._to_info(key) for key in keys]

    def _load_keys(self, path, keys, marker=None):
        headers = {'X-Amz-Prefix': path}
        headers.update({'X-Amz-Marker': marker} if marker is not None else {})

        raw = self.request('GET', '/', headers=headers).text
        response = parse_bucket_listing(raw)
        keys.extend(self._infos_from_keys(response['contents']))

        return keys, response['truncated'], response.get('next', None)

    def findall(self, path=''):
        # Initial loading
        keys, truncated, marker = self._load_keys(path, keys=[])

        # As long as there is more to load, do it.
        while truncated:
            keys, truncated, marker = self._load_keys(path, keys, marker)

        return keys

    def list(self, path=''):
        return [info.path for info in self.findall(path)]

    def find(self, obj):
        r = self.request('HEAD', '/' + obj)

        if r.status_code == 200:
            url = 's3://{0}/{1}'.format(self.bucket, obj)
            path = obj
            mtime = parsedate(r.headers['Last-Modified'])
            etag = r.headers['ETag'].strip('"')
            size = int(r.headers['Content-Length'])
            return interface.ObjectInfo(url, path, mtime, size, etag, {}, self)
        elif r.status_code == 404:
            raise exceptions.ObjectDoesNotExistException.construct(self.bucket, obj)
        else:
            raise exceptions.LuggageException.from_respone(r)

    def delete(self, obj):
        r = self.request('DELETE', '/' + obj)

        if r.status_code == 204:
            return True
        elif r.status_code == 404:
            return False
        else:
            raise exceptions.LuggageException.from_respone(r)

    def put_file(self, local_file, remote_path=None, verify=False, content_type='application/octet-stream'):
        # TODO: Implement verify switch
        # TODO: Implement adding metadata to the put operation
        # TODO: A TON of other stuff...

        # TODO: Oh, that and the code duplication...
        remote_path = remote_path or basename(local_file)
        r = self.request('PUT', remote_path, data=open(local_file), headers={
            'Content-Type': content_type
        })

        if r.status_code / 100 != 2:
            raise exceptions.LuggageException.from_respone(r)

    def put_data(self, data, remote_path, verify=False, content_type='application/octet-stream'):
        r = self.request('PUT', remote_path, data=data, headers={
            'Content-Type': content_type
        })

        if r.status_code / 100 != 2:
            raise exceptions.LuggageException.from_respone(r)

    def get(self, remote_path, local_file=None):
        if local_file:
            destination = abspath(local_file)
        else:
            destination = abspath(remote_path)

        if '/' in destination:
            destination_directory = split(destination)[0]
            if not exists(destination_directory):
                makedirs(destination_directory)

        r = self.request('GET', remote_path, stream=True)

        if r.status_code == 200:
            with open(destination, 'w+') as f:
                for chunk in r.iter_content(4096):
                    f.write(chunk)
            return destination
        else:
            raise exceptions.LuggageException.from_respone(r)

    def read(self, obj, binary=False):
        r = self.request('GET', obj)

        if r.status_code == 200:
            return r.text if binary else r.content
        elif r.status_code == 404:
            raise exceptions.ObjectDoesNotExistException.construct(self.bucket, obj)
        else:
            raise exceptions.LuggageException.from_respone(r)

    def exists(self, obj):
        r = self.request('HEAD', obj)

        if r.status_code / 100 == 2:
            return True
        elif r.status_code == 404:
            return False
        else:
            raise exceptions.LuggageException.from_respone(r)


class ContainerDoesNotExistException(Exception):
    pass


class ObjectDoesNotExistException(Exception):
    @classmethod
    def construct(cls, container, object):
        return cls('Object {0} does not exist in container {1}'.format(
            container, object
        ))


class InvalidCredentialsException(Exception):
    pass


class LuggageException(Exception):
    @classmethod
    def from_respone(cls, response):
        return cls('Unexpected HTTP response - '
                                'status code: {0}, content: "{1}"'.format(
            response.status_code, response.text
        ))


import hashlib
import hmac
from datetime import datetime
from time import mktime
from wsgiref.handlers import format_date_time


def md5(data):
    """
    Calculates the MD5 of the given data or an empty string if the data is not set
    """
    if data:
        if isinstance(data, file):
            data = data.read()
        return hashlib.md5(data).hexdigest()
    else:
        return ''


def rfc1123_date(date=None):
    """
    Formats a date object as required for the HTTP Date field.

    See: http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html
    """
    date = date or datetime.now()
    timestamp = mktime(date.timetuple())
    return format_date_time(timestamp)


def hmac_sha1(key, to_sign):
    """
    Quick & simple HMAC_S
    """
    return hmac.new(key, to_sign, digestmod=hashlib.sha1).digest()
